from typing import Optional

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response, Requests


logger = get_logger(__name__)


def move_job(lifecycle_url: Optional[str], job_name: str, job_version: str, new_infra_target: str):
    """Move job from one infrastructure target to another"""
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.info(f'Moving job "{job_name}" v{job_version} at {lifecycle_url} to {new_infra_target} infrastructure...')
    r = Requests.post(
        f'{lifecycle_url}/api/v1/job/{job_name}/{job_version}/move',
        json={
            'infrastructure_target': new_infra_target,
        },
        headers=get_auth_request_headers(user_auth),
    )
    parse_response(r, 'Lifecycle response error')
    logger.info(f'Job "{job_name}" v{job_version} at {lifecycle_url} has been moved to {new_infra_target} infrastructure ')



import json
import logging
import os
from datetime import datetime
import time

import sys
from typing import Optional, Dict, Any

from racetrack_client.utils.env import is_env_flag_enabled
from racetrack_client.utils.strings import strip_ansi_colors
from racetrack_client.utils.time import timestamp_to_iso8601

LOG_FORMAT = '\033[2m[%(asctime)s]\033[0m %(levelname)s %(message)s'
LOG_FORMAT_DEBUG = '\033[2m[%(asctime)s]\033[0m %(levelname)s %(name)s:%(lineno)s %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
ISO_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

structured_logs_on: bool = is_env_flag_enabled('LOG_STRUCTURED', 'false')
log_caller_enabled: bool = is_env_flag_enabled('LOG_CALLER_NAME', 'false')
debug_format_enabled: bool = is_env_flag_enabled('LOG_FORMAT_DEBUG', 'false')

logger: logging.Logger = logging.getLogger('racetrack')


def configure_logs(log_level: Optional[str] = None):
    """Configure root logger with a log level"""
    log_level = log_level or os.environ.get('LOG_LEVEL', 'debug')
    level = _parse_logging_level(log_level)
    # Set root level to INFO to avoid printing a ton of garbage DEBUG logs from imported libraries
    log_format = LOG_FORMAT_DEBUG if debug_format_enabled else LOG_FORMAT
    logging.basicConfig(stream=sys.stdout, format=log_format, level=logging.INFO, datefmt=LOG_DATE_FORMAT, force=True)

    original_formatter = logging.getLogger().handlers[0].formatter
    if structured_logs_on:
        formatter = StructuredFormatter()
    else:
        formatter = ColoredFormatter(original_formatter)
    set_logging_formatter(formatter)

    root_logger = logging.getLogger('racetrack')
    root_logger.setLevel(level)


def get_logger(_logger_name: str) -> logging.Logger:
    """Get configured racetrack logger"""
    return logging.getLogger('racetrack').getChild(_logger_name)


def set_logging_formatter(formatter: logging.Formatter):
    for handler in logging.getLogger().handlers:
        handler.setFormatter(formatter)


def _format_extra_vars(extra: Dict[str, Any]) -> str:
    if len(extra) == 0:
        return ''
    keys = extra.keys()
    parts = [_format_extra_var(key, extra[key]) for key in keys]
    return ', '.join(parts)


def _format_extra_var(var: str, val: Any) -> str:
    val = str(val)
    if ' ' in val:
        return f'{var}="{val}"'
    else:
        return f'{var}={val}'


def _parse_logging_level(str_level: str) -> int:
    return {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARNING,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'crit': logging.CRITICAL,
        'off': logging.NOTSET,
    }[str_level.lower()]


class ColoredFormatter(logging.Formatter):
    def __init__(self, plain_formatter):
        logging.Formatter.__init__(self)
        self.plain_formatter = plain_formatter

    log_level_templates = {
        'CRITICAL': '\033[1;31mCRIT \033[0m',
        'ERROR': '\033[1;31mERROR\033[0m',
        'WARNING': '\033[0;33mWARN \033[0m',
        'INFO': '\033[0;34mINFO \033[0m',
        'DEBUG': '\033[0;32mDEBUG\033[0m',
    }

    def format(self, record: logging.LogRecord) -> str:
        part_time = self.format_time()
        part_levelname = self.log_level_templates.get(record.levelname, record.levelname)
        log_message: str = record.getMessage()
        logger_details = self._logger_details(record)
        if logger_details:
            line = f'{part_time} {part_levelname} {logger_details}: {log_message}'
        else:
            line = f'{part_time} {part_levelname} {log_message}'

        record_dict = record.__dict__
        extra: Dict[str, Any] = record_dict.get('extra') or {}
        if record_dict.get('cause'):
            extra['cause'] = record_dict.get('cause')
        if record_dict.get('traceback'):
            extra['traceback'] = record_dict.get('traceback')
        if record_dict.get('tracing_id'):
            extra['tracing_id'] = record_dict.get('tracing_id')
        if record_dict.get('caller_name'):
            extra['caller_name'] = record_dict.get('caller_name')
        if extra:
            line += ', ' + _format_extra_vars(extra)

        if not sys.stdout.isatty():
            line = strip_ansi_colors(line)
        return line

    def format_time(self) -> str:
        now_tz = datetime.now().astimezone()
        if time.timezone == 0:
            time_formatted = now_tz.strftime('%Y-%m-%d %H:%M:%SZ')
        else:
            time_formatted = now_tz.strftime('%Y-%m-%d %H:%M:%S')
        return f'\033[2m[{time_formatted}]\033[0m'

    def _logger_details(self, record: logging.LogRecord) -> Optional[str]:
        logger_name = record.name
        if debug_format_enabled or not logger_name.startswith('racetrack'):
            return f'{logger_name}:{record.lineno}'
        else:
            return None


class StructuredFormatter(logging.Formatter):
    def __init__(self, **kwargs):
        logging.Formatter.__init__(self)

    def format(self, record: logging.LogRecord) -> str:
        time_str: str = timestamp_to_iso8601(record.created)
        level: str = record.levelname
        message: str = strip_ansi_colors(record.msg)

        record_dict = record.__dict__
        extra: Dict[str, Any] = record_dict.get('extra') or {}

        log_rec = {
            "time": time_str,
            "lvl": level,
            "msg": message,
            **extra,
        }

        if record_dict.get('cause'):
            log_rec['cause'] = record_dict.get('cause')
        if record_dict.get('traceback'):
            log_rec['traceback'] = record_dict.get('traceback')
        if record_dict.get('tracing_id'):
            log_rec['tracing_id'] = record_dict.get('tracing_id')
        if record_dict.get('caller_name'):
            log_rec['caller_name'] = record_dict.get('caller_name')

        if record.exc_info and record.exc_text is None:
            record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            log_rec["exc_info"] = record.exc_text
        if record.stack_info:
            log_rec["stack_info"] = self.formatStack(record.stack_info)

        return json.dumps(log_rec)



import json
from typing import Dict, Optional

import yaml

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.context_error import wrap_context
from racetrack_client.log.logs import configure_logs, get_logger
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response, parse_response_object, Requests

logger = get_logger(__name__)


def call_job(
    name: str,
    version: str,
    remote: Optional[str],
    endpoint: str,
    payload: str,
    curl: bool,
):
    if curl:
        configure_logs(log_level='error')

    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    user_auth = get_user_auth(client_config, lifecycle_url)

    r = Requests.get(
        f'{lifecycle_url}/api/v1/job/{name}/{version}',
        headers=get_auth_request_headers(user_auth),
    )
    job = parse_response_object(r, 'Lifecycle response error')
    pub_url = job.get('pub_url')
    if not endpoint.startswith('/'):
        endpoint = '/' + endpoint
    full_url = f'{pub_url}{endpoint}'
    payload_dict = _parse_payload(payload)

    if curl:
        _print_curl_command(full_url, payload_dict, user_auth)
        return

    logger.info(f'Calling job "{name} {version}" at {full_url}')
    r = Requests.post(
        full_url,
        json=payload_dict,
        headers=get_auth_request_headers(user_auth),
    )

    response_object = parse_response(r, 'Lifecycle response error')
    logger.info('Response:')
    response_str = json.dumps(response_object, indent=4)
    print(response_str)


def _parse_payload(payload: str) -> Dict:
    with wrap_context('parsing payload'):
        try:
            obj = json.loads(payload)
        except json.JSONDecodeError:
            obj = yaml.load(payload, Loader=yaml.FullLoader)
        assert isinstance(obj, dict), 'given payload is not a key-value dictionary'
        return obj
        

def _print_curl_command(
    full_url: str,
    payload_dict: Dict,
    user_auth: str,
):
    payload_json = json.dumps(payload_dict, indent=2)
    print(f"""
    curl -X 'POST' \\
  '{full_url}' \\
  -H 'X-Racetrack-Auth: {user_auth}' \\
  -H 'Accept: application/json' \\
  -H 'Content-Type: application/json; charset=utf-8' \\
  -d '{payload_json}'
""".strip())



from urllib.parse import urlsplit


def trim_url(url: str) -> str:
    """Trim trailing slash from the end of URL address"""
    split = urlsplit(url)
    path = split.path
    while path.endswith('/'):
        path = path[:-1]
    split = split._replace(path=path)
    return split.geturl()


def join_paths(*paths: str) -> str:
    """Join paths trimming trailing slashes"""
    parts = []
    for i, path in enumerate(paths):
        if i > 0:
            while path.startswith('/'):
                path = path[1:]
        while path.endswith('/'):
            path = path[:-1]
        if path:
            parts.append(path)
    return '/'.join(parts)



from typing import List


def print_table(table: List[List[str]]):
    """
    Print padded table to stdout
    :param table: list of rows (row -> column -> cell)
    """
    cols_num: int = max([len(row) for row in table])

    col_widths: List[int] = []
    for col_index in range(cols_num):
        max_width = 0
        for row in table:
            cell = row[col_index] or ''
            max_width = max(max_width, len(cell))
        col_widths.append(max_width)

    for row in table:
        row_values: List[str] = []
        for col_index in range(cols_num):
            padding = col_widths[col_index]
            cell = (row[col_index] or '').ljust(padding)
            row_values.append(cell)
        print('   '.join(row_values))



from typing import Optional
from getpass import getpass

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.client_config import ClientConfig
from racetrack_client.client_config.io import load_client_config, save_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.auth import is_auth_required, validate_user_auth, AuthError
from racetrack_client.utils.request import Requests, parse_response_object

logger = get_logger(__name__)


def login_user_auth(lifecycle_url: Optional[str], user_auth: str):
    client_config: ClientConfig = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    username = validate_user_auth(lifecycle_url, user_auth)
    set_user_auth(client_config, lifecycle_url, user_auth)
    save_client_config(client_config)
    logger.info(f'Logged as user {username} to Racetrack: {lifecycle_url}')


def logout_user_auth(lifecycle_url: Optional[str]):
    client_config: ClientConfig = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    set_user_auth(client_config, lifecycle_url, "")
    save_client_config(client_config)
    logger.info(f'Logged out from Racetrack: {lifecycle_url}')


def set_user_auth(client_config: ClientConfig, lifecycle_url: str, user_auth: str):
    """
    You need to save the resulting config if you want to make changes persist
    """
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)

    if len(user_auth) == 0:
        if lifecycle_url in client_config.user_auths:
            del client_config.user_auths[lifecycle_url]
        else:
            raise RuntimeError(f'Missing {lifecycle_url} in Racetrack logged servers')
    else:
        client_config.user_auths[lifecycle_url] = user_auth


def get_user_auth(client_config: ClientConfig, lifecycle_url: str) -> str:
    if lifecycle_url in client_config.user_auths:
        return client_config.user_auths[lifecycle_url]

    if is_auth_required(lifecycle_url):
        raise AuthError(f"missing auth token for {lifecycle_url}. You need to do: racetrack login <token> --remote {lifecycle_url}")

    return ''


def get_current_auth(remote: Optional[str]) -> str:
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    return get_user_auth(client_config, lifecycle_url)


def login_with_username(remote: Optional[str], username: str):
    client_config = load_client_config()
    lifecycle_url = resolve_lifecycle_url(client_config, remote)
    password = getpass(f'Enter your password for user {username}: ')
    auth_token = authorize_username_password(lifecycle_url, username, password)
    set_user_auth(client_config, lifecycle_url, auth_token)
    save_client_config(client_config)
    logger.info(f'Logged as user {username} to Racetrack: {lifecycle_url}')


def authorize_username_password(lifecycle_url: str, username: str, password: str) -> str:
    response = Requests.post(f'{lifecycle_url}/api/v1/users/login', json={
        'username': username,
        'password': password,
    })
    if response.status_code == 401:
        raise AuthError('Incorrect username or password')
    response_dict = parse_response_object(response, 'Lifecycle response error')
    return response_dict['token']



from typing import Optional, List

from pydantic import BaseModel, ConfigDict


class PluginManifest(BaseModel):
    """Details of the contents of the plugin"""
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True)

    name: str

    version: str

    build_date: Optional[str] = None
    
    # URL of the plugin homepage
    url: Optional[str] = None

    # order in plugins sequence, lowest priority gets executed first
    priority: int = 0

    # kind of the plugin, e.g. 'infrastructure', 'job-type' or 'core'
    category: Optional[str] = None

    # list of Racetrack components that the plugin should be running on, e.g. 'lifecycle', 'image-builder'
    components: List[str] = []

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))



from typing import Dict, Optional, List

import backoff

from racetrack_client.manifest import Manifest
from racetrack_client.utils.shell import CommandError, shell
from racetrack_client.utils.time import datetime_to_timestamp, now
from racetrack_client.client.deploy import DEPLOYMENT_TIMEOUT_SECS, BuildContextMethod, DeploymentError, get_build_context, get_deploy_request_payload, get_git_credentials
from racetrack_client.client_config.io import load_client_config
from racetrack_client.manifest.load import get_manifest_path
from racetrack_client.manifest.merge import load_merged_manifest_dict
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.log.logs import get_logger
from racetrack_client.client.env import hide_env_vars, merge_env_vars, read_secret_vars
from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response_object, Requests

logger = get_logger(__name__)

JOB_INTERNAL_PORT = 7000  # Job listening port seen from inside the container


def run_job_locally(
    workdir: str,
    lifecycle_url: str,
    build_context_method: BuildContextMethod = BuildContextMethod.default,
    port: Optional[int] = None,
    extra_vars: Dict[str, str] = None,
    build_flags: List[str] = [],
    cmd: Optional[str] = None,
):
    client_config = load_client_config()
    manifest: Manifest = load_validated_manifest(workdir, extra_vars)
    manifest_dict: Dict = load_merged_manifest_dict(get_manifest_path(workdir), extra_vars)

    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.info(f'building workspace "{workdir}" (job {manifest.name} v{manifest.version}) by {lifecycle_url}')

    build_context = get_build_context(lifecycle_url, workdir, manifest, build_context_method)
    git_credentials = get_git_credentials(manifest, client_config)
    secret_vars = read_secret_vars(workdir, manifest)

    # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::BuildEndpoint` for server-side implementation
    r = Requests.post(
        f'{lifecycle_url}/api/v1/build',
        json=get_deploy_request_payload(manifest_dict, git_credentials, secret_vars, build_context, False, build_flags),
        headers=get_auth_request_headers(user_auth),
    )
    response = parse_response_object(r, 'Lifecycle building error')
    deployment_id = response["id"]
    logger.info(f'job buidling requested: {deployment_id}')

    try:
        job_image_name = _wait_for_building_result(lifecycle_url, deployment_id, user_auth)
    except Exception as e:
        raise DeploymentError(e)

    logger.info(f'Job image "{manifest.name}" v{manifest.version} has been built.')

    try:
        shell(f'docker', print_stdout=False)
    except CommandError as e:
        raise RuntimeError('Docker is not installed in the system') from e

    docker_registry = extract_docker_registry(job_image_name)
    logger.info(f'Logging in to Docker registry: {docker_registry}')
    shell(f'docker login {docker_registry}', raw_output=True)

    logger.info(f'Pulling image {job_image_name}')
    shell(f'docker pull {job_image_name}', raw_output=True)

    port = port or 7000
    container_name = f'local-job-{manifest.name}-v-{manifest.version}'.replace('.', '-')

    deployment_timestamp = datetime_to_timestamp(now())
    runtime_env_vars = merge_env_vars(manifest.runtime_env, secret_vars.runtime_env)
    build_env_vars = merge_env_vars(manifest.build_env, secret_vars.build_env)
    runtime_env_vars = hide_env_vars(runtime_env_vars, build_env_vars)
    common_env_vars: Dict[str, str] = {
        'JOB_NAME': manifest.name,
        'JOB_DEPLOYMENT_TIMESTAMP': str(deployment_timestamp),
    }
    runtime_env_vars = merge_env_vars(runtime_env_vars, common_env_vars)
    env_vars_cmd = ' '.join([f'--env {env_name}="{env_val}"' for env_name, env_val in runtime_env_vars.items()])

    job_url = f'http://127.0.0.1:{port}/pub/job/{manifest.name}/{manifest.version}/'
    cmd = cmd or ''
    logger.info(f'Running job "{manifest.name}" v{manifest.version} locally. '
                f'Check out {job_url} to access your job. CTRL-C to stop.')
    try:
        shell(
            f'docker run --rm -it'
            f' --name {container_name}'
            f' -p {port}:{JOB_INTERNAL_PORT}'
            f' {env_vars_cmd}'
            f' --label job-name={manifest.name}'
            f' --label job-version={manifest.version}'
            f' {job_image_name}'
            f' {cmd}'
        )
    except CommandError as e:
        if e.returncode == 130:  # Container terminated by Control-C
            return
        raise e


@backoff.on_exception(
    backoff.fibo, TimeoutError, max_value=5, max_time=DEPLOYMENT_TIMEOUT_SECS, jitter=None, logger=None
)
def _wait_for_building_result(lifecycle_url: str, deployment_id: str, user_auth: str) -> str:
    # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::DeployIdEndpoint` for server-side implementation
    r = Requests.get(
        f'{lifecycle_url}/api/v1/deploy/{deployment_id}',
        headers=get_auth_request_headers(user_auth),
    )
    response = parse_response_object(r, 'Lifecycle building status')
    status = response['status'].lower()
    if status == 'failed':
        raise RuntimeError(response['error'])
    if status == 'done':
        return response['image_name']

    phase = response.get('phase')
    if phase:
        logger.info(f'building in progress: {phase}...')
    else:
        logger.info(f'building in progress...')
    raise TimeoutError('Building timeout error')


def extract_docker_registry(job_image_name: str) -> str:
    return job_image_name.split('/')[0]
